/*
   ESP32C3_BLE_Scanner_ServerTCP_UDP_NTP.cpp
   by Matthew Ford,  2021/12/06
   (c)2021 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.

*/

// Needs Huge APP, 2M APP, 1M SPIFF  Partition setting

#include <BLEDevice.h>
#include <BLEUtils.h>
#include <BLEScan.h>
#include <BLEAdvertisedDevice.h>

#include <WiFiClient.h>
#include <WebServer.h>
#include "ntpSupport.h"

#define DEBUG
static Stream *debugPtr = NULL;

WebServer server(80);
void startWebServer();
void startTelnetServer();
void handleTelnetConnection();

#include "ESPAutoWiFiConfig.h"  // download ESPAutoWiFiConfig.zip from  https://www.forward.com.au/pfod/ESPAutoWiFiConfig/index.html
#include "pfodLinkedPointerList.h" // iterable linked list of pointers to objects
#include "LastSeen.h"  // class for storing on linked list
#include "SafeString.h"
#include <WiFi.h>

// settings for the Auto WiFi connect
bool highForLedOn = true; // turns led OFF when WiFi connected
//and, -ve indicates the ws2812 driver is to be used
int ledPin = -8; // for ESP32-C3 ws2818 pin
size_t eepromOffset = 40; // if you use EEPROM.begin(size) in your code add the size here so AutoWiFi data is written after your data
// need to add ntpSupport getTimeZoneStorageSize() to this offset

pfodLinkedPointerList<LastSeen> listOfLastSeen;

// returns NULL if not found
LastSeen* getLastSeen(SafeString &name) {
  LastSeen *devicePtr = listOfLastSeen.getFirst();
  while (devicePtr) {
    if (name.startsWith(devicePtr->getDeviceName())) {
      return devicePtr;
    } // else
    devicePtr = listOfLastSeen.getNext();
  }
  return devicePtr; // null not found
}

int scanTime = 2; //In seconds
BLEScan *pBLEScan;

class MyAdvertisedDeviceCallbacks : public BLEAdvertisedDeviceCallbacks {
    void onResult(BLEAdvertisedDevice advertisedDevice) {
      if (advertisedDevice.haveName()) {
        cSF(sfName, 50); // max 32
        sfName = advertisedDevice.getName().c_str();
        if (debugPtr) {
          debugPtr->print("Device name: ");
          debugPtr->println(sfName);
        }
        // only scan for == upto first ,
        int idx = sfName.indexOf(',');
        if (idx >= 0) {
          sfName.substring(sfName, 0, idx);
          sfName += ',';
        }

        LastSeen *devicePtr = getLastSeen(sfName);
        if (!devicePtr) {
          // not  found add it upto first ,
          if (debugPtr) {
            debugPtr->print("Adding Device name: "); debugPtr->println(sfName);
          }
          devicePtr = new LastSeen(sfName.c_str()); // note MUST use new since pfodLinkedPointerList uses delete when remove() called
          listOfLastSeen.add(devicePtr);
        }
        // update lastseen
        devicePtr->updateLastSeen(millis());
        devicePtr->setAdvertisedName(advertisedDevice.getName().c_str()); // save the full name
      }
    }
};

static TaskHandle_t bleScannerHandle = NULL;


void BLE_init() {
  BLEDevice::init("");
  pBLEScan = BLEDevice::getScan(); //create new scan
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setActiveScan(true);
  pBLEScan->setInterval(40);  // same as Apple in foreground
  pBLEScan->setWindow(30); // same as Apple in foreground, less or equal setInterval value
}

/* this function will be invoked when additionalTask was created */
void bleScannerTask( void * parameter ) {
  BLE_init();
  /* loop forever */
  for (;;) {
    // put your taskain code here, to run repeatedly:
    pBLEScan->clearResults(); // delete results fromBLEScan buffer to release memory
    BLEScanResults foundDevices = pBLEScan->start(scanTime, false);
    if (debugPtr) {
      debugPtr->print("Devices found: ");
      debugPtr->println(foundDevices.getCount());
    }
    yield();
  }
  /* delete a task when finish,
    this will never happen because this is infinity loop */
  vTaskDelete( NULL );
}

void setUpWiFiServices() {
  setNtpSupportDebug(debugPtr);
  initializeNtpSupport();
  resetDefaultTZstr(); // only need this first time through
  startWebServer();
  startTelnetServer();
}
void setup() {
  Serial.begin(115200);

#ifdef DEBUG
  SafeString::setOutput(Serial); // for debug error msgs
  debugPtr = &Serial;
  debugPtr->println();
#endif
  /* Initialise wifi module */
#ifdef DEBUG
  setESPAutoWiFiConfigDebugOut(Serial); // turns on debug output for the ESPAutoWiFiConfig code
#endif

  if (ESPAutoWiFiConfigSetup(ledPin, highForLedOn, eepromOffset)) { // check if we should start access point to configure WiFi settings
    return; // in config mode so skip rest of setup
  }

  setUpWiFiServices(); // do this first. Get continual reboots if create BLE task first and then call this
  BaseType_t err = xTaskCreate(
                     bleScannerTask, /* Task function. */
                     "bleScannerTask", /* name of task. */
                     102400, /* Stack size of task */
                     NULL, /* parameter of the task */
                     1, /* priority of the task */ // the loop task also has priority 1 also so will share
                     &bleScannerHandle); /* Task handle to keep track of created task */
  if (((int)err) != 1) { // pPass ,  pdTRUE      ( ( BaseType_t ) 1 ) in #include "freertos/projdefs.h"
    if (debugPtr) {
      debugPtr->print("xTaskCreatePinnedToCore returned:");
      debugPtr->println((int)err);
    }
  }
}



void loop() {
  //  .. other code that MUST run all the time
  if (ESPAutoWiFiConfigLoop()) {  // handle WiFi config webpages
    return;  // skip the rest of the loop until config finished
  }
  server.handleClient();
  yield();
  processNTP();
  yield();
  handleTelnetConnection();
  yield();
}

// ------------------ trival telnet server -------------------------
//how many clients should be able to telnet to this ESP32
#define MAX_SRV_CLIENTS 1

WiFiServer telnetServer(23);
WiFiClient telnetServerClients[MAX_SRV_CLIENTS];

void startTelnetServer() {
  telnetServer.begin();
  telnetServer.setNoDelay(true);
  if (debugPtr) {
    debugPtr->print("Ready! Use 'telnet ");
    debugPtr->print(WiFi.localIP());
    debugPtr->println(" 23' to connect");
  }
}

void handleTelnetConnection() {
  //check if there are any new clients
  uint8_t i;
  if (telnetServer.hasClient()) {
    for (i = 0; i < MAX_SRV_CLIENTS; i++) {
      //find free/disconnected spot
      if (!telnetServerClients[i] || !telnetServerClients[i].connected()) {
        if (telnetServerClients[i]) telnetServerClients[i].stop();
        telnetServerClients[i] = telnetServer.available();
        if (!telnetServerClients[i]) {
          if (debugPtr) {
            debugPtr->println("available broken");
          }
        }
        if (debugPtr) {
          debugPtr->print("New client: ");
          debugPtr->print(i); debugPtr->print(' ');
          debugPtr->println(telnetServerClients[i].remoteIP());
        }
        break;
      }
    }
    if (i >= MAX_SRV_CLIENTS) {
      //no free/disconnected spot so reject
      telnetServer.available().stop();
    }
  }
  //check clients for data
  for (i = 0; i < MAX_SRV_CLIENTS; i++) {
    if (telnetServerClients[i] && telnetServerClients[i].connected()) {
      if (telnetServerClients[i].available()) {
        //get data from the telnet client and push it to the UART
        Serial.println();
        Serial.print(" >>>> Telnet:");
        while (telnetServerClients[i].available()) Serial.write(telnetServerClients[i].read());
        Serial.println();
      }
    }
    else {
      if (telnetServerClients[i]) {
        telnetServerClients[i].stop();
      }
    }
  }
  //check UART for data
  if (Serial.available()) {
    size_t len = Serial.available();
    uint8_t sbuf[len];
    Serial.readBytes(sbuf, len);
    //push UART data to all connected telnet clients
    for (i = 0; i < MAX_SRV_CLIENTS; i++) {
      if (telnetServerClients[i] && telnetServerClients[i].connected()) {
        telnetServerClients[i].write(sbuf, len);
        delay(1);
      }
    }
  }
}


// ========== trivial webserver example =============
void notFound() {
  server.send(404, "text/plain", "Not found");
}

void handleRoot() {
  if (debugPtr) {
    debugPtr->println(">>> WebServer handleRoot");
  }
  String msg =
    "<html>\
  <head>\
    <meta http-equiv='refresh' content='5'/>\
    <title>BLE Temperature Sensors</title>\
    <style>\
      body { background-color: #cccccc; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }\
    </style>\
  </head>\
  <body>";
  time_t now;
  msg += "TimeZone: ";
  msg += getCurrentTZ();
  msg += "<br>";
  msg += getCurrentTZdescription();
  msg += "<p>At ";
  now = time(nullptr); // nullptr is a C++ null pointer literal which you can use anywhere you need to pass a null pointer.
  msg += ctime(&now);
  msg += "<br>";
  msg += "The BLE devices found were:-<br>";
  // loop through and add reading found
  LastSeen *devicePtr = listOfLastSeen.getFirst();
  msg += "<h1>";
  if (devicePtr == NULL) {
    msg += "No devices found so far.<br>";
  } else {
    while (devicePtr != NULL) {
      msg += devicePtr->getAdvertisedName();
      msg += "<font size=\"-1\"> ";
      msg += (millis() - devicePtr->getLastSeen()) / 1000.0;
      msg += " sec ago</font>";
      msg += "<br>";
      devicePtr = listOfLastSeen.getNext();
    }
  }
  msg += "</h1>";
  msg += "</body></html>";

  server.send(200, "text/html", msg);
}

void startWebServer() {
  server.on("/", handleRoot);
  server.onNotFound(notFound);
  server.begin();
  if (debugPtr) {
    debugPtr->println("webserver started");
  }
}
